/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * "Portions Copyright (c) 1999 Apple Computer, Inc.  All Rights
 * Reserved.  This file contains Original Code and/or Modifications of
 * Original Code as defined in and that are subject to the Apple Public
 * Source License Version 1.0 (the 'License').  You may not use this file
 * except in compliance with the License.  Please obtain a copy of the
 * License at http://www.apple.com/publicsource and read it before using
 * this file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License."
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
#define	K_ADD	1
	{"add", K_ADD},
#define	K_BLACKHOLE	2
	{"blackhole", K_BLACKHOLE},
#define	K_CHANGE	3
	{"change", K_CHANGE},
#define	K_CLONING	4
	{"cloning", K_CLONING},
#define	K_DELETE	5
	{"delete", K_DELETE},
#define	K_DST	6
	{"dst", K_DST},
#define	K_EXPIRE	7
	{"expire", K_EXPIRE},
#define	K_FLUSH	8
	{"flush", K_FLUSH},
#define	K_GATEWAY	9
	{"gateway", K_GATEWAY},
#define	K_GENMASK	10
	{"genmask", K_GENMASK},
#define	K_GET	11
	{"get", K_GET},
#define	K_HOST	12
	{"host", K_HOST},
#define	K_HOPCOUNT	13
	{"hopcount", K_HOPCOUNT},
#define	K_IFACE	14
	{"iface", K_IFACE},
#define	K_INTERFACE	15
	{"interface", K_INTERFACE},
#define	K_IFA	16
	{"ifa", K_IFA},
#define	K_IFP	17
	{"ifp", K_IFP},
#define	K_INET	18
	{"inet", K_INET},
#define	K_INET6	19
	{"inet6", K_INET6},
#define	K_ISO	20
	{"iso", K_ISO},
#define	K_LINK	21
	{"link", K_LINK},
#define	K_LLINFO	22
	{"llinfo", K_LLINFO},
#define	K_LOCK	23
	{"lock", K_LOCK},
#define	K_LOCKREST	24
	{"lockrest", K_LOCKREST},
#define	K_MASK	25
	{"mask", K_MASK},
#define	K_MONITOR	26
	{"monitor", K_MONITOR},
#define	K_MTU	27
	{"mtu", K_MTU},
#define	K_NET	28
	{"net", K_NET},
#define	K_NETMASK	29
	{"netmask", K_NETMASK},
#define	K_NOSTATIC	30
	{"nostatic", K_NOSTATIC},
#define	K_OSI	31
	{"osi", K_OSI},
#define	K_PREFIXLEN	32
	{"prefixlen", K_PREFIXLEN},
#define	K_PROTO1	33
	{"proto1", K_PROTO1},
#define	K_PROTO2	34
	{"proto2", K_PROTO2},
#define	K_RECVPIPE	35
	{"recvpipe", K_RECVPIPE},
#define	K_REJECT	36
	{"reject", K_REJECT},
#define	K_RTT	37
	{"rtt", K_RTT},
#define	K_RTTVAR	38
	{"rttvar", K_RTTVAR},
#define	K_SA	39
	{"sa", K_SA},
#define	K_SENDPIPE	40
	{"sendpipe", K_SENDPIPE},
#define	K_SSTHRESH	41
	{"ssthresh", K_SSTHRESH},
#define	K_STATIC	42
	{"static", K_STATIC},
#define	K_X25	43
	{"x25", K_X25},
#define	K_XNS	44
	{"xns", K_XNS},
#define	K_XRESOLVE	45
	{"xresolve", K_XRESOLVE},
